/* ---------------------------------------------------------------------- *\
  dump-rx.c

  PCI register dump for the TT-RX module (fifo only)

  Eunil Won, Korea University

  $Id: test-ttrx_fifo.c,v 1.1.1.1 2005/02/23 09:44:12 ttrx Exp $
  $Log: test-ttrx_fifo.c,v $
  Revision 1.1.1.1  2005/02/23 09:44:12  ttrx
  First into cvs


  2004091301  re-written for the devive driver 
\* ---------------------------------------------------------------------- */

#define VERSION 2005011500
#define TTRX_MAX_LENGTH 0x450
#define TTRX_FIFO_DEV_HEAD "/dev/ttrx_fifo"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "ttrxlib.h"

/* ---------------------------------------------------------------------- *\
   regdump
\* ---------------------------------------------------------------------- */
void
regdump(int *base, int readfifo, int summaryonly)
{
  unsigned int pid     = base[TTRX_PID]   & 0xf;
  unsigned int pver    = base[TTRX_PVER]  & 0xf;
  unsigned int prgm    = base[TTRX_PRGM]  & 0xf;
  unsigned int sclk    = base[TTRX_SCLK]  & 0xf;

  unsigned int rclk    = base[TTRX_RCLK];
  unsigned int id      = base[TTRX_ID];
  unsigned int ver     = base[TTRX_VER];
  unsigned int creg    = base[TTRX_CREG];
  unsigned int freg    = base[TTRX_FREG];
  unsigned int intr    = base[TTRX_INTR];

  unsigned int trig    = base[TTRX_TRIG];
  unsigned int tlast   = base[TTRX_TLAST];
  unsigned int rate    = base[TTRX_RATE];
  unsigned int tag     = base[TTRX_TAG];
  unsigned int tagr    = base[TTRX_TAGR];
  unsigned int depth   = base[TTRX_DEPTH];

  unsigned int tio     = base[TTRX_TIO];
  unsigned int ttest   = base[TTRX_TTEST];
  unsigned int tstat   = base[TTRX_TSTAT];
  unsigned int tsav    = base[TTRX_TSAV];
  unsigned int tsav2   = base[TTRX_TSAV2];
  unsigned int tng32   = base[TTRX_TNG32];

  char buf[256];
  double freq;
  int i;
  
  if (! summaryonly) {
    printf("------------------------------------");
    printf("------------------------------------\n");
    printf("dump-ttrx version %d\n", VERSION);
    printf("------------------------------------");
    printf("------------------------------------\n");
    printf("CPLD:\n");
    printf("000    ID=%01x %s\n", pid,  pid==0x3?"OK":(pid==2?"OK(v3)":"NG"));
    printf("010   VER=%01x (CPLD version 0.%02d)\n", pver, pver);
    printf("020  PRGM=%01x (lock*=%d m012=%d init=%d done=%d)\n",
	   prgm, (prgm>>3)&1, (prgm>>2)&1, (prgm>>1)&1, prgm&1);
    printf("030  SCLK=%01x (tclksel=%d clk=%s edgei=%d edgeo=%d)\n",
	   sclk, (sclk>>3)&1, sclk&4?"global":"local", (sclk>>1)&1, sclk&1);
    
    printf("\n");
    printf("FPGA:\n");
    printf("080  RCLK=%08x rdiv=%d ttrig=%d localtrg=%d trig=%d trigin=%d ext=%x\n",
	   rclk, rclk&0xff,
	   (rclk>>8)&1, (rclk>>9)&1, (rclk>>10)&1, (rclk>>11)&1,
	   (rclk>>12)&15);
    printf("100    ID=%08x %s\n", id, id==0x58525454?"OK":"NG");
    printf("110   VER=%08x TTRX version %d FPGA version %d.%02d rev %02d\n",
	   ver, (ver>>28)&0xf, (ver>>16)&0xfff, (ver>>8)&0xff, ver&0xff);
    printf("120  CREG=%08x revo=%c frst=%c srst=%c lrst=%c iena=%c abrt=%c\n",
	   creg,
	   (creg>>31)&1?(creg>>23)&1+'0':'Z',
	   (creg>>30)&1?(creg>>22)&1+'0':'Z',
	   (creg>>29)&1?(creg>>21)&1+'0':'Z',
	   (creg>>28)&1?(creg>>20)&1+'0':'Z',
	   (creg>>27)&1?(creg>>19)&1+'0':'Z',
	   (creg>>26)&1?(creg>>18)&1+'0':'Z');
    printf("%19sentestbsy=%d enful=%d exbsy=%x bsydone=%x ready=%d\n",
	   "", (creg>>17)&1, (creg>>16)&1,
	   (creg>>12)&15, (creg>>8)&15, (creg>>7)&1);
    printf("%19snwff=%d testbsy=%d swbsy=%d orun=%d full=%d half=%d emp=%d\n",
	   "", (creg>>6)&1, (creg>>5)&1, (creg>>4)&1,
	   (creg>>3)&1, (creg>>2)&1, (creg>>1)&1, (creg>>0)&1);
    
    printf("130  FREG=%08x io2/io1/ful/emp/bsy/hand/flg/mask:DCBA\n", freg);
    printf("140  INTR=%08x irq=%d encdbg=%d\n", intr, intr&1, (intr>>31)&1);
    printf("200  TRIG=%08x disg=%d decg=%d enau=%d pls=%d rnd=%d free=%d n=%d\n",
	   trig, (trig>>31)&1, (trig>>30)&1, (trig>>29)&1,
	   (trig>>28)&1, (trig>>27)&1, (trig>>26)&1, trig&((1<<24)-1));
    printf("210 TLAST=%08x n=%d\n", tlast, tlast);
    printf("220  RATE=%08x refclk=%d Hz[28:24] rndset=%d[17:16] plsset=1/%d[9:0]\n",
	   rate,
	   42333000/(1<<((rate>>24)&31)),
	   (rate>>16)&3,
	   (rate&((1<<10)-1))+1);
    printf("230   TAG=%08x\n", tag);
    printf("250 DEPTH=%08x\n", depth);
    if (readfifo) {
      unsigned int fifo1   = base[0x30 << 2];
      unsigned int fifo2   = base[0x30 << 2 + 1];
      printf("\n");
      printf("300   fifo1=%08x\n", fifo1);
      printf("304   fifo2=%08x\n", fifo2);
    }
    printf("400   TIO=%08x in=%03x out=%03x usr=%03x\n",
	   tio, (tio>>20)&0x3ff, (tio>>10)&0x3ff, tio&0x3ff);
    printf("410 TTEST=%08x othru=%d oincr=%d enusr=%d\n",
	   ttest, (ttest>>30)&1, (ttest>>29)&1, (ttest>>28)&1);
    printf("%19stcmd=%1x tout=%02x tin_de=%02x tdiff=%02x\n",
	   "", (ttest>>24)&15, (ttest>>16)&255, (ttest>>8)&255, ttest&255);
    printf("420 TSTAT=%08x Nng1=%d Nng2=%d Nngc=%d Nngany=%d\n",
	   tstat, (tstat>>24)&255, (tstat>>16)&255, (tstat>>8)&255, tstat&255);
    printf("430  TSAV=%08x ng1=%d ng2=%d x1=%d x2=%d x=%d ngc=%d\n",
	   tsav, (tsav>>31)&1, (tsav>>30)&1, (tsav>>29)&1, (tsav>>28)&1,
	   (tsav>>27)&1, (tsav>>24)&1);
    printf("%19stincr=%02x tin=%02x texp=%02x\n",
	   "", (tsav>>16)&255, (tsav>>8)&255, tsav&255);
    printf("440 TSAV2=%08x ng1=%d ng2=%d x1=%d x2=%d x=%d ngc=%d\n",
	   tsav2, (tsav2>>31)&1, (tsav2>>30)&1, (tsav2>>29)&1, (tsav2>>28)&1,
	   (tsav2>>27)&1, (tsav2>>24)&1);
    printf("%19stincr=%02x tin=%02x texp=%02x\n",
	   "", (tsav2>>16)&255, (tsav2>>8)&255, tsav2&255);
    printf("450 TNG32=%08x count=%d\n", tng32, tng32);
    printf("------------------------------------");
    printf("------------------------------------\n");
  }

  if (1) {
    printf("Summary:\n");
    buf[0] = 0;
    freq = 0;
    if (trig & (1<<28)) {
      sprintf(buf+strlen(buf), "%s%s", *buf?"+":"", "Pulse");
      freq = 42333000.0/(1<<((rate>>24)&31))/(1+rate&((1<<10)-1));
    }
    if (trig & (1<<27)) {
      sprintf(buf+strlen(buf), "%s%s", *buf?"+":"", "Random");
      freq = 42333000.0/(1<<((rate>>24)&31))/512;
    }
    if (trig & (1<<29)) sprintf(buf+strlen(buf), "%s%s", *buf?"+":"", "User");
    if (! *buf) strcpy(buf, "None");
    
    printf("    Trigger source=%s  count=%d", buf, tag);
    
    if (freq > 0) printf(" freq=%3.1f Hz", freq);
    if (trig & ((1<<24)-1)) {
      printf(" out of %d (last=%d)\n", trig & ((1<<24)-1), tlast);
    } else {
      printf(" (free-run)\n");
    }
    
    printf("  Handshake ");
    if (creg & (1<<7)) {
      printf("READY\n");
    } else {
      int rx_fifofull = 0;
      
      printf("NOT-READY");
      creg &= ((freg>>8)&15)<<12 | ~(15<<12);
      if (creg & (15<<12)) {
	printf(" extra-busy=%s%s%s%s",
	       (creg & (1<<12)) ? "A" : "",
	       (creg & (2<<12)) ? "B" : "",
	       (creg & (4<<12)) ? "C" : "",
	       (creg & (8<<12)) ? "D" : "");
      }
      freg &= ~((freg&15)<<20);      /* mask => full */
      freg &= ~((freg&15)<<12);      /* mask => busy */
      freg &= ~(((freg>>12)&15)<<8); /* busy => hand (for nobusy) */
      freg &= ~(((freg>>4)&15)<<12); /* flag => busy */
      if (freg & (15<<8)) {
	printf(" no-busy=%s%s%s%s",
	       (freg & (1<<8)) ? "A" : "",
	       (freg & (2<<8)) ? "B" : "",
	       (freg & (4<<8)) ? "C" : "",
	       (freg & (8<<8)) ? "D" : "");
      }
      if (freg & (15<<12)) {
	printf(" busy=%s%s%s%s",
	       (freg & (1<<12)) ? "A" : "",
	       (freg & (2<<12)) ? "B" : "",
	       (freg & (4<<12)) ? "C" : "",
	       (freg & (8<<12)) ? "D" : "");
      }
      if ((creg & (1<<2)) && ! (creg & (1<<25))) rx_fifofull++;
      if ((creg & (1<<1)) && ! (creg & (1<<24))) rx_fifofull++;
      
      if ((freg & (15<<20)) || (creg & (1<<6)) || rx_fifofull) {
	printf(" fifo-full=%s%s%s%s%s%s",
	       (freg & (1<<20)) ? "A" : "",
	       (freg & (2<<20)) ? "B" : "",
	       (freg & (4<<20)) ? "C" : "",
	       (freg & (8<<20)) ? "D" : "",
	       (creg & (1<<6))  ? "(NWFF)" : "",
	       rx_fifofull ? "(TT-RX)" : "");
      }
      printf("\n");
    }
  }

  if (! summaryonly) {
    printf("------------------------------------");
    printf("------------------------------------\n");
  }
}
/* ---------------------------------------------------------------------- *\
   main
\* ---------------------------------------------------------------------- */
int
main(int argc, char **argv)
{
  unsigned int base_plx;
  unsigned int base_rx = 0;
  unsigned int *ptr_plx;
  unsigned int *ptr_rx;
  int size_plx  = 256;
  int size_rx = 4096;
  int readfifo = 0;
  int summaryonly = 0;
  int rxid = 0;

  int fd;
  int nread, nwritten; 
  int fifo[2];
  int regs;
  int base_new[((TTRX_MAX_LENGTH)>>2)+1];
  char ttrx_dev_name[64];

  while (argc > 1) {
    if (strncmp(argv[1], "-b", 2) == 0) {
      base_rx = strtoul(argv[2], 0, 16);
      argc--, argv++;
    } else if (strncmp(argv[1], "-f", 2) == 0) {
      readfifo = ! readfifo;
    } else if (strncmp(argv[1], "-s", 2) == 0) {
      summaryonly = ! summaryonly;
    } else if (argv[1][0] == '-' && isdigit(argv[1][1])) {
      rxid = argv[1][1] - '0';
    } else {
      printf("usage: dump-rx [-<n>] [-f] [-b <base>]\n");
      printf("options:\n");
      printf(" -<n>       select <n>-th TT-RX (<n>=0,1,2...9)\n");
      printf(" -f           read fifo\n");
      printf(" -s           summary only\n");
      printf(" -b <base>    set base address (in hex)\n");
      return 1;
    }
    argc--, argv++;
  }

  sprintf(ttrx_dev_name,TTRX_FIFO_DEV_HEAD"%d",rxid);

  if ((fd = open(ttrx_dev_name,O_RDWR)) < 0) 
  {
    printf("TT-RX is not found.\n");
    exit(1);
  }

  nread = read(fd,fifo,8);
  printf("0x%04x fifo1 = 0x%08x \n",TTRX_FIFO1,fifo[0]);
  printf("fifo1 addr 0x%08x \n",&fifo[0]);
  printf("0x%04x fifo2 = 0x%08x \n",TTRX_FIFO2,fifo[1]);
  printf("fifo2 addr 0x%08x \n",&fifo[1]);


/*
  lseek(fd,TTRX_ID,SEEK_SET);
  regs = 0xdeafbeef;
  nwritten = write(fd, &regs, 4);
*/
  close(fd);

/*
  ptr_rx = base;
  regdump(ptr_rx, readfifo, summaryonly);
*/
  return 0;
}
