/* ---------------------------------------------------------------------- *\
   xfer-ttrx.c

   Serial bus test - this program assumes two TT-RX on the same PCI bus,
   interconnected with the special TT-RXv2 cross cable.

  Mikihiko Nakao, KEK IPNS
  Eunil Won, Korea University

  $Id: xfer-ttrx.c,v 1.1.1.1 2005/02/23 09:44:12 ttrx Exp $
  $Log: xfer-ttrx.c,v $
  Revision 1.1.1.1  2005/02/23 09:44:12  ttrx
  First into cvs


  2005011500  re-written for the devive driver 

\* ---------------------------------------------------------------------- */

#define VERSION 2005011500
#define TTRX_MAX_LENGTH 0x450
#define TTRX_DEV_HEAD "/dev/ttrx"

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <sys/time.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "ttrxlib.h"

int ptn[256][256];
time_t t0;
int n, m, count, counted = 0, unit = 1;


/* ---------------------------------------------------------------------- *\
   graycode
\* ---------------------------------------------------------------------- */
int
graycode(int prev)
{
  static int nextone[256];
  if (! nextone[0]) {
    int seq16[] = { 0, 1, 3, 2, 6, 7, 5, 4, 12, 13, 15, 14, 10, 11, 9, 8 };
    int seq[256];
    int i, j;

    for (i=0; i<16; i++) {
      for (j=0; j<16; j++) {
        if (i & 1) {
          seq[i*16 + j] = seq16[i]*16 + seq16[15-j];
        } else {
          seq[i*16 + j] = seq16[i]*16 + seq16[j];
        }
      }
    }
    for (i=0; i<256; i++) {
      j = (i + 1) & 0xff;
      nextone[seq[i]] = seq[j];
    }
  }
  return nextone[prev];
}
/* ---------------------------------------------------------------------- *\
   handler
\* ---------------------------------------------------------------------- */
void
handler(int sig)
{
  int np[256];
  int i, j;
  int sum = 0;
  time_t t1 = time(0);
  struct tm *tmp = localtime(&t1);

  for (i=0; i<256; i++) {
    int find = 0;
    for (j=0; j<256; j++) {
      if (ptn[i][j]) break;
    }
    if (j == 256) continue;
    if (sig >= 0) printf("expect=%02x:", i);
    for (j=0; j<256; j++) {
      if (sig >= 0 && ptn[i][j]) {
        printf(" %02x*%d", j, ptn[i][j]);
        if (j == graycode(i)) {
          printf("+");
        } else if (i == graycode(j)) {
          printf("-");
        }
      }
      sum += ptn[i][j];
    }
    if (sig >= 0) printf("\n");
  }

  printf("%02d:%02d:%02d (%d sec) count=%d*%d+%d %d errors\n",
                 tmp->tm_hour, tmp->tm_min, tmp->tm_sec,
                 t1 - t0, counted, unit, n, sum);

  if (sig >= 0) {
    printf("exit.\n");
    exit(0);
  }
}
/* ---------------------------------------------------------------------- *\
   main
\* ---------------------------------------------------------------------- */
int
main(int argc, char **argv)
{
  volatile unsigned int *ptr_rx1, *ptr_rx2;
  unsigned int base_plx1, base_plx2, base_rx1, base_rx2;
  int size_rx = 4096;
  time_t t1;
  int num = 0;
  int prgm, tio, tout, expect, prev;
  int i, count;
  int counts[100];
  double loop = 0;

  int fd_1, fd_2;
  int nread, nwritten;
  int base_1, base_2;
  char ttrx_dev_name[64];

  memset(counts, 0, sizeof(counts));

  printf("xfer-ttrx version %d\n", VERSION);

  t0 = time(0);
  t1 = t0 + 300;
  memset(ptn, 0, sizeof(ptn));
  signal(SIGINT, handler);

  /*
   * device to write
   */
  sprintf(ttrx_dev_name,TTRX_DEV_HEAD"%d",1);
  if ((fd_1 = open(ttrx_dev_name,O_RDWR)) < 0) 
  {
    printf("TT-RX is not found.\n");
    exit(1);
  }

  /*
   * device to read 
   */
  sprintf(ttrx_dev_name,TTRX_DEV_HEAD"%d",0);
  if ((fd_2 = open(ttrx_dev_name,O_RDONLY)) < 0) 
  {
    printf("TT-RX is not found.\n");
    exit(1);
  }

  lseek(fd_1,TTRX_TTEST,SEEK_SET); 
  base_1 = (1<<28); nwritten = write(fd_1,&base_1,4);
  lseek(fd_1,TTRX_TIO,SEEK_SET); 
  base_1 = 0x000;   nwritten = write(fd_1,&base_1,4);
  lseek(fd_1,TTRX_TIO,SEEK_SET); 
  base_1 = 0x2aa;   nwritten = write(fd_1,&base_1,4);

  usleep(10000); 
  lseek(fd_2,TTRX_PRGM,SEEK_SET);
  nread = read(fd_2,&prgm,4);
  lseek(fd_2,TTRX_TIO,SEEK_SET);
  nread = read(fd_2,&tio, 4);
  tout = (tio>>20) & 0x3ff;
  if ((prgm & 0x8) || tout != 0x2aa) {
    printf("cannot sync TT-RX 1 => 2 prgm=%01x tio=%08x (tout=%03x)\n",
           prgm, tio, tout);
    return -1;
  }

 lseek(fd_1,TTRX_TIO,SEEK_SET); 
 lseek(fd_2,TTRX_TIO,SEEK_SET); 

  prev = 0x2aa<<20;
  t1 = t0 + 10;
  while (1) {
    for (i=0; i<1000; i++) {
      count = 0;
      expect = num << 20;
      nwritten = write(fd_1,&num,4); 
      while (1) {
        nread= read (fd_2,&tio,4);
        tout = tio & (0x3ff<<20);
        if (tout != prev || ++count == 100) break;
      }
      if (tout != expect) {
        printf("bad tio=%08x (tout=%03x) expect=%08x prev=%08x count=%d in=%08x\n",
               tio, tout, expect, prev, count, ptr_rx1[TTRX_TIO]);
        return -1;
      }
      if (count == 100) {
        printf("cannot update within 100 loop cycle\n");
        return -1;
      }
      counts[count]++;
      num = graycode(num);
      prev = expect;

      if (time(0) >= t1) {
        printf("loop=%1.0f num=%02x rate=%1.0fHz count=%d/%d/%d t=%ds\n",
               (loop+i), num, loop/(t1-t0)/1000.0,
               counts[0], counts[1], counts[2], t1-t0);
        t1 += 10;
      }
    }
    loop += 1000;
  }
}
